//+------------------------------------------------------------------+
//|                                                      ProjectName |
//|                                      Copyright 2012, CompanyName |
//|                                       http://www.companyname.net |
//+------------------------------------------------------------------+
#ifndef _C_LIST_
#define _C_LIST_

#include <CObject.mqh>

#define INSERT                                                   \
   mPtr.SetIteratorNext(cPtr.GetNextIterator());                 \
   mPtr.SetIteratorPrev(cPtr.GetPrevIterator());                 \
   if(!IsFirst()) cPtr.GetPrevIterator().SetIteratorNext(mPtr);  \
   if(!IsLast()) cPtr.GetNextIterator().SetIteratorPrev(mPtr);   \
cPtr.ResetObjIterators();
//---------------------------------------------------------------------------------------------------   
#define CHECK_LAST_END                \
   if(!(cIterator>0)) cBegine=cPtr;   \
if(cIterator+1==cSize) cEnd=cPtr;
//---------------------------------------------------------------------------------------------------
#define CHECK_DELETE                                                                              \
   if(!IsFirst()) cPtr.GetPrevIterator().SetIteratorNext(IsLast()?NULL:cPtr.GetNextIterator());   \
   if(!IsLast()) cPtr.GetNextIterator().SetIteratorPrev(IsFirst()?NULL:cPtr.GetPrevIterator());   \
   if(--cSize==cIterator){                                                                        \
      cPtr=cPtr.GetPrevIterator();                                                                \
      --cIterator;}                                                                               \
   else cPtr=cPtr.GetNextIterator();                                                              \
   if(cIterator<0) cEnd=cBegine=NULL;                                                             \
   else {CHECK_LAST_END}                                                                           

template<typename T>
class CList:public CObject
  {
   T                *cPtr;
   T                *cBegine;
   T                *cEnd;
   int               cIterator;
   int               cSize;
   bool              cIsDeletable;
private:
   T                *Get(int mPos);
public:
                     CList():cPtr(NULL),cBegine(NULL),cEnd(NULL),cIterator(-1),cSize(0),cIsDeletable(true){}
                    ~CList() {if(cIsDeletable) DeleteAll();}
   int               GetIterator()  {return cIterator;}
   int               GetSize()      {return cSize;}
   bool              IsEmpty()      {return !cSize;}
   bool              IsFirst()      {return cIterator<=0;}
   bool              IsLast()       {return cIterator+1==cSize;}
   bool              IsDeletable() {return cIsDeletable;}
   T                *Push(T *mPtr,int mPos=INT_MIN);
   T                *PushBack(T *mPtr);
   T                *PushFirst(T* mPtr)   {return !cSize?PushBack(mPtr):Push(mPtr,0);}
   T                *PushNext(T* mPtr)    {return !cSize||IsLast()?PushBack(mPtr):Push(mPtr,cIterator+1);}
   T                *Swap(T *mPtr,int mPos=INT_MIN);
   T                *Insert(T *mPtr,int mPos=INT_MIN);
   T                *Erace(int mPos=INT_MIN);
   T                *PopFirst()  {return Move(0);}
   T                *PopBack()   {return Move(cSize-1);}
   T                *It()     {return cPtr;}
   T                *End()    {if(cSize) cIterator=cSize-1; return cPtr=cEnd;}
   T                *Begine() {if(cSize) cIterator=0; return cPtr=cBegine;}
   T                *Next(int mCount=1);
   T                *Prev(int mCount=1);
   T                *Move(int mPos=INT_MIN);
   T                *Delete(int mPos=INT_MIN);
   void              DeleteAll() {while(Delete()!=NULL);}
   void              SetDestructMode(bool mIsDeletable) {cIsDeletable=mIsDeletable;}
   T *operator[](int mPos) {return Get(mPos);}
  };
//-----------------------------------------------------
template<typename T>
T *CList::Push(T *mPtr,int mPos=INT_MIN)
  {
   if(mPos==INT_MIN) mPos=cIterator;
   else if(Get(mPos)==NULL) return NULL;
   if(cSize)
     {
      if(!IsFirst())
        {
         cPtr.GetPrevIterator().SetIteratorNext(mPtr);
         mPtr.SetIteratorPrev(cPtr.GetPrevIterator());
        }
      mPtr.SetIteratorNext(cPtr);
      cPtr.SetIteratorPrev(mPtr);
     }
   cPtr=mPtr;
   if(cIterator<0)++cIterator;
   ++cSize;
   CHECK_LAST_END
   return cPtr;
  }
//----------------------------------------------------
template<typename T>
T *CList::PushBack(T *mPtr)
  {
   if(cSize)
     {
      End().SetIteratorNext(mPtr);
      mPtr.SetIteratorPrev(cPtr);
      mPtr.SetIteratorNext(NULL);
     }
   cPtr=mPtr;
   ++cIterator;
   ++cSize;
   cEnd=mPtr;
   if(!cIterator) cBegine=mPtr;
   return cPtr;
  }
//----------------------------------------------------
template<typename T>
T *CList::Next(int mCount=1)
  {
   if(mCount<0||!cSize||cSize-cIterator<=mCount) return NULL;
   int stop=cIterator+mCount;
   while(stop>cIterator) {cPtr=cPtr.GetNextIterator();++cIterator;}
   return cPtr;
  }
//----------------------------------------------------
template<typename T>
T *CList::Prev(int mCount=1)
  {
   if(mCount<0||cIterator<mCount) return NULL;
   int stop=cIterator-mCount;
   while(stop<cIterator) {cPtr=cPtr.GetPrevIterator(); --cIterator;}
   return cPtr;
  }
//------------------------------------------------------
template<typename T>
T *CList::Move(int mPos=INT_MIN)
  {
   if(mPos==INT_MIN) mPos=cIterator;
   else if(Get(mPos)==NULL) return NULL;
   if(cPtr==NULL) return NULL;
   T *ptr=cPtr;;
   CHECK_DELETE
   ptr.ResetObjIterators();
   return ptr;
  }
//------------------------------------------------------
template<typename T>
T *CList::Swap(T *mPtr,int mPos=INT_MIN)
  {
   if(mPos==INT_MIN) mPos=cIterator;
   else if(Get(mPos)==NULL) return NULL;
   if(cPtr==NULL) {PushBack(mPtr); return NULL;}
   T *ptr=cPtr;
   INSERT
   cPtr=mPtr;
   CHECK_LAST_END
   return ptr;
  }
//------------------------------------------------------
template<typename T>
T *CList::Insert(T *mPtr,int mPos=INT_MIN)
  {
   if(mPos==INT_MIN) mPos=cIterator;
   else if(Get(mPos)==NULL) return NULL;
   if(cPtr==NULL) return PushBack(mPtr);
   INSERT
   delete cPtr;
   cPtr=mPtr;
   CHECK_LAST_END
   return cPtr;
  }
//------------------------------------------------------
template<typename T>
T *CList::Erace(int mPos=INT_MIN)
  {
   if(mPos==INT_MIN) mPos=cIterator;
   else if(Get(mPos)==NULL) return NULL;
   if(cPtr==NULL) return NULL;
   CHECK_DELETE
   return cPtr;
  }
//------------------------------------------------------
template<typename T>
T *CList::Delete(int mPos=INT_MIN)
  {
   if(mPos==INT_MIN) mPos=cIterator;
   else if(Get(mPos)==NULL) return NULL;
   if(cPtr==NULL) return NULL;
   T *ptr=(IsLast())?cPtr.GetPrevIterator():cPtr.GetNextIterator();
   delete cPtr;
   cPtr=ptr;
   if(--cSize==cIterator) --cIterator;
   CHECK_LAST_END
   return cPtr;
  }
//------------------------------------------------------
template<typename T>
T *CList::Get(int mPos)
  {
   if(mPos<0||mPos>=cSize) return NULL;
   return mPos==cIterator?cPtr:mPos>cIterator?Next(mPos-cIterator):Prev(cIterator-mPos);
  }
//------------------------------------------------------

#undef INSERT
#undef CHECK_LAST_END
#undef CHECK_DELETE

#endif 
//+------------------------------------------------------------------+
